/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/** gclc:net.bigeon.gclc.ConsoleApplication.java
 * Created on: Sep 6, 2014 */
package net.bigeon.gclc;

/*-
 * #%L
 * Generic Command Ligne console
 * %%
 * Copyright (C) 2014 - 2018 bigeon.fr
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.bigeon.gclc.command.ICommand;
import net.bigeon.gclc.command.ICommandProvider;
import net.bigeon.gclc.command.SubedCommand;
import net.bigeon.gclc.exception.CommandParsingException;
import net.bigeon.gclc.exception.CommandRunException;
import net.bigeon.gclc.exception.CommandRunExceptionType;
import net.bigeon.gclc.exception.InvalidCommandName;
import net.bigeon.gclc.i18n.Messages;
import net.bigeon.gclc.manager.ConsoleInput;
import net.bigeon.gclc.manager.ConsoleOutput;

/**
 * A {@link ConsoleApplication} is an application that require the user to input
 * commands.
 * <p>
 * A typical use case is the following:
 *
 * <pre>
 * {@link ConsoleOutput} out = new {@link net.bigeon.gclc.utils.StreamConsoleOutput StreamConsoleOutput}();
 * {@link ConsoleInput} in = new {@link net.bigeon.gclc.utils.StreamConsoleInput StreamConsoleInput}();
 * {@link ConsoleApplication} app = new {@link ConsoleApplication}(out, in, "welcome", "see you latter")};
 * app.{@link ConsoleApplication#add(ICommand) add}("my_command", new {@link ICommand MyCommand()});
 * app.{@link ConsoleApplication#start() start()};
 * </pre>
 * <p>
 * That will launch in the console application that will display "welcome",
 * return a line and prompt the user for the command until he inputs the
 * <code>exit</code> command. Then it will print "see you latter" and exit the
 * start method.
 *
 * @author Emmanuel BIGEON */
public final class ConsoleApplication implements ICommandProvider {

    /** The class logger. */
    private static final Logger LOGGER = Logger
            .getLogger(ConsoleApplication.class.getName());
    /** The welcome message. */
    public final String header;
    /** The good bye message. */
    public final String footer;
    /** The standard output for the application. */
    private final ConsoleOutput out;
    /** The standard input for the application. */
    private final ConsoleInput in;
    /** The container of commands. */
    public final SubedCommand root;
    /** The state of this application. */
    private boolean running;
    /** The listeners. */
    private final List<CommandRequestListener> listeners = new ArrayList<>();

    /** Create a console application.
     *
     * @param out the output
     * @param in the input
     * @param welcome the welcoming message
     * @param goodbye the goodbye message */
    public ConsoleApplication(final ConsoleOutput out, final ConsoleInput in,
            final String welcome,
            final String goodbye) {
        header = welcome;
        footer = goodbye;
        this.in = in;
        this.out = out;
        root = new SubedCommand(""); //$NON-NLS-1$
    }

    @Override
    public boolean add(final ICommand cmd) throws InvalidCommandName {
        return root.add(cmd);
    }

    /** Add a listener for command requests.
     *
     * @param listener the command listener */
    public void addListener(final CommandRequestListener listener) {
        listeners.add(listener);
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.command.ICommandProvider#executeSub(java.lang.String,
     * java.lang.String[]) */
    @Override
    public void executeSub(final ConsoleOutput output, final ConsoleInput input,
                           final String command,
                                 final String... args) throws CommandRunException {
        root.executeSub(output, input, command, args);
    }

    /** Signify to the application that no command should be inputed anymore. */
    public void exit() {
        LOGGER.fine("Request exiting application..."); //$NON-NLS-1$
        running = false;
        in.interruptPrompt();
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.command.ICommandProvider#get(java.lang.String) */
    @Override
    public ICommand get(final String command) {
        return root.get(command);
    }

    /** Interpret a command line.
     * <p>
     * This method will split the command in its part and execute the command
     * with {@link #executeSub(ConsoleOutput, ConsoleInput, String, String...)}.
     *
     * @param cmd the command
     * @throws IOException if the command could not be parsed */
    public void interpretCommand(final String cmd) throws IOException {
        List<String> args;
        try {
            args = GCLCConstants.splitCommand(cmd);
        } catch (final CommandParsingException e1) {
            out.println("Command line cannot be parsed"); //$NON-NLS-1$
            LOGGER.log(Level.FINE, "Invalid user command " + cmd, e1); //$NON-NLS-1$
            return;
        }
        if (!args.isEmpty()) {
            try {
                executeSub(out, in, args.get(0), Arrays.copyOfRange(
                        args.toArray(new String[0]), 1, args.size()));
            } catch (final CommandRunException e) {
                LOGGER.log(Level.FINE, "Command failed: " + cmd, e); //$NON-NLS-1$
                out.println(Messages
                        .getString("ConsoleApplication.cmd.failed", cmd)); //$NON-NLS-1$
                out.println(e.getLocalizedMessage());
                if (e.getType() == CommandRunExceptionType.USAGE) {
                    get(args.get(0)).help(out);
                }
            }
        }
    }

    /** Test if the application is running.
     *
     * @return the running status */
    public boolean isRunning() {
        return running;
    }

    /** Remove a listener from this application.
     *
     * @param listener the command listener to remove */
    public void removeListener(final CommandRequestListener listener) {
        listeners.remove(listener);
    }

    /** The running loop content.
     * <p>
     * This consisting in getting the command, executing it and exiting
     * (restarting the loop). */
    private void runLoop() {
        try {
            final String cmd = in.prompt();
            if (cmd == null || cmd.isEmpty()) {
                return;
            }
            for (final CommandRequestListener listener : listeners) {
                listener.commandRequest(cmd);
            }
            interpretCommand(cmd);
        } catch (final InterruptedIOException e) {
            LOGGER.info(
                    "Prompt interrupted. It is likely the application is closing."); //$NON-NLS-1$
            LOGGER.log(Level.FINER, "Interruption of the prompt.", //$NON-NLS-1$
                    e);
        } catch (final IOException e) {
            // The manager was closed
            running = false;
            LOGGER.warning(
                    "The console manager was closed. Closing the application as no one can reach it."); //$NON-NLS-1$
            LOGGER.log(Level.FINE,
                    "An exception caused the closing of the application", //$NON-NLS-1$
                    e);
        }
    }

    /** Start the application. */
    public void start() {
        try {
            running = true;
            if (header != null) {
                out.println(header);
            }
        } catch (final IOException e) {
            // The manager was closed
            running = false;
            LOGGER.warning(
                    "The console manager was closed. Closing the application as no one can reach it."); //$NON-NLS-1$
            LOGGER.log(Level.FINE,
                    "An exception caused the closing of the application", //$NON-NLS-1$
                    e);
            return;
        }
        do {
            runLoop();
        } while (running);
        if (footer != null) {
            try {
                out.println(footer);
            } catch (final IOException e) {
                // The manager was closed
                running = false;
                LOGGER.warning("Console manager already closed."); //$NON-NLS-1$
                LOGGER.log(Level.FINE,
                        "Exception raised by goodbye message printing... Application will still close.", //$NON-NLS-1$
                        e);
            }
        }
        running = false;
        LOGGER.fine("Exiting application."); //$NON-NLS-1$
    }
}
