/** acide:fr.bigeon.acide.CommandProvider.java
 * Created on: Aug 6, 2014 */
package net.bigeon.gclc.command;

/*-
 * #%L
 * Generic Command Ligne console
 * %%
 * Copyright (C) 2014 - 2018 bigeon.fr
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.util.LinkedHashSet;
import java.util.Set;

import net.bigeon.gclc.exception.CommandRunException;
import net.bigeon.gclc.exception.InvalidCommandName;
import net.bigeon.gclc.i18n.Messages;
import net.bigeon.gclc.manager.ConsoleInput;
import net.bigeon.gclc.manager.ConsoleOutput;

/** A command provider is a map of key word to command to execute.
 *
 * @author Emmanuel BIGEON */
public class CommandProvider implements ICommandProvider {
    /** The minus character. */
    private static final String   MINUS = "-"; //$NON-NLS-1$
    /** The space character. */
    private static final String   SPACE = " "; //$NON-NLS-1$
    /** The commands set.
     * <p>
     * The insertion order is conserved through the use of a
     * {@link LinkedHashSet}. */
    protected final Set<ICommand> commands;

    /** Create a command provider. */
    public CommandProvider() {
        super();
        commands = new LinkedHashSet<>();
    }

    /** Test the command name validity.
     *
     * @param name the command name
     * @throws InvalidCommandName if the name is invalid */
    private static void testCommandName(final String name) throws InvalidCommandName {
        if (name == null || name.isEmpty() || name.startsWith(MINUS)
                || name.contains(SPACE)) {
            throw new InvalidCommandName();
        }
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.command.ICommandProvider#add(java.lang.String,
     * net.bigeon.gclc.command.Command) */
    @Override
    public final boolean add(final ICommand value) throws InvalidCommandName {
        final String name = value.getCommandName();
        testCommandName(name);
        if (commands.contains(value)) {
            return true;
        }
        for (final ICommand iCommand : commands) {
            if (iCommand.getCommandName().equals(value.getCommandName())) {
                throw new InvalidCommandName(
                        "Name already used: " + value.getCommandName()); //$NON-NLS-1$
            }
        }
        return commands.add(value);
    }

    /* (non-Javadoc)
     * @see
     * net.bigeon.gclc.command.ICommandProvider#executeSub(net.bigeon.gclc.manager
     * .ConsoleOutput, net.bigeon.gclc.manager.ConsoleInput, java.lang.String,
     * java.lang.String[]) */
    @Override
    public final void executeSub(final ConsoleOutput out, final ConsoleInput in,
            final String cmd, final String... args) throws CommandRunException {
        final ICommand command = get(cmd);
        if (command == null) {
            throw new CommandRunException(
                    Messages.getString("CommandProvider.unrecognized", cmd), null); //$NON-NLS-1$
        }
        command.execute(out, in, args);
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.command.ICommandProvider#get(java.lang.String) */
    @Override
    public final ICommand get(final String commandName) {
        for (final ICommand command : commands) {
            if (command.getCommandName().equals(commandName)) {
                return command;
            }
        }
        return null;
    }
}
